/* Copyright (C) 2013 Timo Sirainen, LGPLv2.1 */

/*
   eval `cat /usr/local/lib/dovecot/dovecot-config`
   gcc -fPIC -shared -DHAVE_CONFIG_H \
     `echo $DOVECOT_CFLAGS $LIBDOVECOT_INCLUDE $LIBDOVECOT_STORAGE_INCLUDE` \
     passdb-openam.c -o passdb_openam.so
*/

#include "auth-common.h"
#include "ioloop.h"
#include "istream.h"
#include "str.h"
#include "uri-util.h"
#include "http-client.h"
#include "http-url.h"
#include "http-response-parser.h"
#include "passdb.h"

#define OPENAM_HTTP_TIMEOUT_MSECS 10000

struct openam_passdb_module {
	struct passdb_module module;
	const char *host, *target_template;
	unsigned int port;
	bool ssl;
};

struct openam_request {
	struct auth_request *auth_request;
	verify_plain_callback_t *callback;

	struct http_client_request *http_req;
	struct istream *payload;
	struct io *io;
	struct timeout *to;
	const char *post_args;
};

static struct http_client *http_client = NULL;

static const char *
http_param_escape(const char *str,
		  const struct auth_request *auth_request ATTR_UNUSED)
{
	string_t *esc = t_str_new(64);

	http_url_escape_param(esc, str);
	return str_c(esc);
}

static void openam_request_free(struct openam_request *request)
{
	if (request->to != NULL)
		timeout_remove(&request->to);
	if (request->io != NULL)
		io_remove(&request->io);
	if (request->payload != NULL)
		i_stream_unref(&request->payload);
	if (request->http_req != NULL)
		http_client_request_abort(&request->http_req);
}

static void openam_request_tempfail(struct openam_request *request)
{
	openam_request_free(request);
	request->callback(PASSDB_RESULT_INTERNAL_FAILURE,
			  request->auth_request);
}

static void payload_input(struct openam_request *request)
{
	const char *line;
	ssize_t ret;

	while ((ret = i_stream_read(request->payload)) > 0) ;
	if (ret == 0)
		return;

	if (request->payload->stream_errno != 0) {
		auth_request_log_error(request->auth_request, "openam",
				       "HTTP request read() failed: %m");
		openam_request_tempfail(request);
		return;
	}

	/* we've read the full HTTP response payload now. parse it through. */
	line = i_stream_read_next_line(request->payload);
	if (line == NULL || strcmp(line, "OK") != 0) {
		auth_request_log_info(request->auth_request, "openam",
				      "Authentication failed");
		openam_request_free(request);
		request->callback(PASSDB_RESULT_PASSWORD_MISMATCH,
				  request->auth_request);
		return;
	}
	openam_request_free(request);
	request->callback(PASSDB_RESULT_OK, request->auth_request);
}

static void
openam_verify_plain_callback(const struct http_response *response,
			     struct openam_request *request)
{
	struct passdb_module *_module = request->auth_request->passdb->passdb;
	struct openam_passdb_module *module =
		(struct openam_passdb_module *)_module;

	request->http_req = NULL;

	if (response->status == 401) {
		auth_request_log_info(request->auth_request, "openam",
				      "Authentication failed");
		openam_request_free(request);
		request->callback(PASSDB_RESULT_PASSWORD_MISMATCH,
				  request->auth_request);
		return;
	}
	if (response->status != 200) {
		auth_request_log_error(request->auth_request, "openam",
				       "HTTP request to %s:%u failed: %s",
				       module->host, module->port,
				       response->reason);
		openam_request_tempfail(request);
		return;
	}

	i_stream_ref(response->payload);
	request->payload = response->payload;
	request->io = io_add(i_stream_get_fd(response->payload), IO_READ,
			     payload_input, request);
	payload_input(request);
}

static void openam_request_timeout(struct openam_request *request)
{
	auth_request_log_error(request->auth_request, "openam",
			       "HTTP request timed out");
	openam_request_tempfail(request);
}

static void openam_http_client_get(const struct auth_settings *set)
{
	struct http_client_settings http_set;

	memset(&http_set, 0, sizeof(http_set));
	//http_set.ssl_ca_dir = set->ssl_client_ca_dir;
	//http_set.ssl_ca_file = set->ssl_client_ca_file;
	http_set.max_parallel_connections = 100;
        //http_set.debug = TRUE;
	//http_set.rawlog_dir = "/tmp/rawlogs";
	http_set.dns_client_socket_path =
		t_strconcat(set->base_dir, "/dns-client", NULL);

	http_client = http_client_init(&http_set);
}

static void
openam_verify_plain(struct auth_request *auth_request,
		    const char *password ATTR_UNUSED,
		    verify_plain_callback_t *callback)
{
	struct passdb_module *_module = auth_request->passdb->passdb;
	struct openam_passdb_module *module =
		(struct openam_passdb_module *)_module;
	struct openam_request *request;
	const struct var_expand_table *vars;
	struct istream *input;
	string_t *path;

	auth_request_log_debug(auth_request, "openam", "lookup");

	if (http_client == NULL)
		openam_http_client_get(auth_request->set);

	request = p_new(auth_request->pool, struct openam_request, 1);
	request->auth_request = auth_request;
	request->callback = callback;

	vars = auth_request_get_var_expand_table(auth_request,
						 http_param_escape);
	path = t_str_new(128);
	var_expand(path, module->target_template, vars);
	request->post_args =
		p_strdup_printf(auth_request->pool, "password=%s",
				http_param_escape(auth_request->mech_password,
						  auth_request));

	request->to = timeout_add(OPENAM_HTTP_TIMEOUT_MSECS,
				  openam_request_timeout, request);
	request->http_req =
		http_client_request(http_client, "POST", module->host,
				    str_c(path), openam_verify_plain_callback,
				    request);
	if (module->port != 0)
		http_client_request_set_port(request->http_req, module->port);
	http_client_request_set_ssl(request->http_req, module->ssl);
	input = i_stream_create_from_data(request->post_args,
					  strlen(request->post_args));
	http_client_request_set_payload(request->http_req, input, FALSE);
	http_client_request_add_header(request->http_req,
				       "Content-Type", "application/x-www-form-urlencoded");
	i_stream_unref(&input);
	http_client_request_submit(request->http_req);
}

static struct passdb_module *
openam_preinit(pool_t pool, const char *args)
{
	struct openam_passdb_module *module;
	struct uri_parser parser;
	struct uri_authority auth;
	const char *scheme;

	module = p_new(pool, struct openam_passdb_module, 1);

	uri_parser_init(&parser, pool_datastack_create(), args);
	if (uri_parse_scheme(&parser, &scheme) <= 0 ||
	    uri_parse_authority(&parser, &auth) <= 0 ||
	    (strcasecmp(scheme, "http") != 0 &&
	     strcasecmp(scheme, "https") != 0))
		i_fatal("passdb openam: Invalid URL: %s", args);

	module->host = p_strdup(pool, auth.host_literal);
	module->ssl = strcasecmp(scheme, "https") == 0;
	if (auth.port != 0)
		module->port = auth.port;
	else if (module->ssl)
		module->port = 443;
	else
		module->port = 80;
	module->target_template = p_strdup_until(pool, parser.cur, parser.end);
	return &module->module;
}

static void openam_deinit(struct passdb_module *module ATTR_UNUSED)
{
}

struct passdb_module_interface passdb_openam = {
	"openam",

	openam_preinit,
	NULL,
	openam_deinit,

	openam_verify_plain,
	NULL,
	NULL
};

void passdb_openam_init(void);
void passdb_openam_deinit(void);

void passdb_openam_init(void)
{
	passdb_register_module(&passdb_openam);
}

void passdb_openam_deinit(void)
{
	passdb_unregister_module(&passdb_openam);
}
